from pydantic import BaseModel, field_validator
from typing import List, Optional
from datetime import date, datetime, time, timedelta
import holidays
import pandas as pd


# Define the model for a period within a day
class Period(BaseModel):
    start_time: str  # e.g., '08:00'
    end_time: str  # e.g., '10:00'
    intensity: float  # Intensity of the activity (0 to 1)
    intensity_number: Optional[int] = 1  # Optional field for intensity number (default is 1)

    @field_validator("intensity")
    def check_intensity(cls, intensity):
        if not (0 <= intensity <= 1):
            raise ValueError("Intensity must be between 0 and 1")
        return intensity


# Define the model for a single day schedule
class DaySchedule(BaseModel):
    day_of_week: int  # 0=Monday, 1=Tuesday, ..., 6=Sunday
    period_list: List[Period]  # List of periods for that day


# Define the model for weekly schedules
class WeekSchedule(BaseModel):
    week_number: int  # Week number in the year (1-52)
    days: List[DaySchedule]  # List of day schedules for that week


# Define the model for holiday periods
class HolidayPeriod(BaseModel):
    start_date: date  # Start date of the holiday
    end_date: date  # End date of the holiday
    holiday_name: str  # Name of the holiday (e.g., 'Christmas')


# Define the main model that holds the entire yearly schedule
class YearSchedule(BaseModel):
    year: int  # The year the schedule applies to
    weekly_schedules: List[WeekSchedule]  # Weekly schedules
    holiday_periods: List[HolidayPeriod]  # List of holiday periods
    additional_notes: Optional[str] = None  # Optional field for any extra notes
    include_public_holidays: bool = False  # Flag to include public holidays
    public_holiday_country: Optional[str] = None  # Country code for public holidays

    @field_validator("year")
    def check_year(cls, year):
        if year < 1000 or year > 9999:
            raise ValueError("Year must be a valid 4-digit year")
        return year

    def add_public_holidays(self):
        if self.include_public_holidays and self.public_holiday_country:
            country_holidays = holidays.CountryHoliday(self.public_holiday_country, years=self.year)
            for holiday_date, holiday_name in country_holidays.items():
                self.holiday_periods.append(
                    HolidayPeriod(start_date=holiday_date, end_date=holiday_date, holiday_name=holiday_name)
                )


def convert_to_series(schedule: YearSchedule) -> pd.Series:
    start_date = pd.Timestamp(f"{schedule.year}-01-01")
    end_date = pd.Timestamp(f"{schedule.year}-12-31")
    index = pd.date_range(start=start_date, end=end_date, freq="15min")
    series = pd.Series(data=0.0, index=index)

    for week_schedule in schedule.weekly_schedules:
        for day_schedule in week_schedule.days:
            for period in day_schedule.period_list:
                week_start = start_date + timedelta(weeks=week_schedule.week_number - 1)
                day_date = week_start + timedelta(days=day_schedule.day_of_week)
                start_time = datetime.combine(day_date, datetime.strptime(period.start_time, "%H:%M").time())
                end_time = datetime.combine(day_date, datetime.strptime(period.end_time, "%H:%M").time())
                series[start_time:end_time] = period.intensity

    for holiday_period in schedule.holiday_periods:
        series[holiday_period.start_date:holiday_period.end_date] = 0.0

    return series


def create_basic_schedule(year: int, country: str = "DK") -> YearSchedule:
    schedule = YearSchedule(year=year, weekly_schedules=[], holiday_periods=[], include_public_holidays=True, public_holiday_country=country)
    schedule.add_public_holidays()

    day_schedule = [DaySchedule(day_of_week=day, period_list=[Period(start_time="08:00", end_time="17:00", intensity=1)]) for day in range(5)]
    for week_number in range(1, 53):
        schedule.weekly_schedules.append(WeekSchedule(week_number=week_number, days=day_schedule))

    return schedule


schedule = create_basic_schedule(2025)
series = convert_to_series(schedule)



# Description: This file contains the TAIL object and the function that assigns the TAIL object to the rooms, sensors and buildings
from .model import Room, Building, Attribute, Indicator, TAIL, TAILRating, Quality
from typing import Dict, List

## CONSTANTS ##
THERMAL = {Attribute.AIR_TEMPERATURE}
ACOUSTIC = {Attribute.SOUND_PRESSURE_LEVEL}
INDOOR_AIR_QUALITY = {
    Attribute.CO2, Attribute.RELATIVE_HUMIDITY, Attribute.BENZENE, Attribute.FORMALDEHYDE, Attribute.PM25, Attribute.RADON, Attribute.VENTILATION_RATE, Attribute.MOLD
}
LIGHTING = {Attribute.ILLUMINANCE, Attribute.DAYLIGHT_FACTOR}

TAIL_INDICATORS = {
    Indicator.THERMAL: THERMAL,
    Indicator.ACOUSTIC: ACOUSTIC,
    Indicator.INDOOR_AIR_QUALITY: INDOOR_AIR_QUALITY,
    Indicator.LIGHTING: LIGHTING
}

## FUNCTIONS ##
def get_tail_rating(room: Room) -> TAIL:
    """
    Get the worst quality for each TAIL indicator in the room.
    """

    for tail_indicator, attributes in TAIL_INDICATORS.items():
        quality = max((room.tail.attributes[attr] for attr in attributes if attr in room.tail.attributes), default=None)
        room.tail.assign_indicator(tail_indicator, quality)

    room.tail.calculate_overall()
    return room

def assign_tail_rating(building: Building) -> Building:
    """
    Assign TAIL rating to each room in the building.
    """
    for room in building.rooms:
        room = get_tail_rating(room)
    return building

def assign_tail_rating_to_building(building: Building) -> Building:
    """
    Take the average of the TAIL ratings of all rooms in the building and assign it to the building.
    """
    if not all(hasattr(room, 'tail') for room in building.rooms):
        raise ValueError("All rooms in the building must have a TAIL rating assigned before assigning to the building")
    
    tail = TAIL()
    attr_quality: Dict[Attribute, List[Quality]] = {}
    indicator_quality: Dict[Indicator, List[Quality]] = {}

    for room in building.rooms:
        for attr in Attribute.list():
            if attr in room.tail.attributes:
                attr_quality.setdefault(attr, []).append(room.tail.attributes[attr])
        for indicator in Indicator.list():
            if indicator in room.tail.indicators:
                indicator_quality.setdefault(indicator, []).append(room.tail.indicators[indicator])

    for attr in Attribute.list():
        tail.attributes[attr] = Quality.average(attr_quality[attr]) if attr in attr_quality else None
    for indicator in Indicator.list():
        tail.indicators[indicator] = Quality.average(indicator_quality[indicator]) if indicator in indicator_quality else None

    tail.calculate_overall()

    from_date = min(room.room_df.index.min() for room in building.rooms)
    to_date = max(room.room_df.index.max() for room in building.rooms)

    building.tail_rating = TAILRating(
        from_date=from_date,
        to_date=to_date,
        tail=tail,
        name=f"{building.name} TAIL Rating from {from_date} to {to_date}",
        building_id=building.id,
        rooms_id=[room.id for room in building.rooms],
        sensors_id=[sensor.id for room in building.rooms for sensor in room.sensors],
        valid=True
    )
        
    return building



import pandas as pd
import numpy as np
from datetime import datetime
from typing import Dict, List

from .model import Building, SensorData, Attribute, BuildingType, Quality, Mold, RoomType
from . import const
from .ventilation import get_ventilation_rates
from .schedule import convert_to_series

def get_quality_for_attr(series: pd.Series, attribute: Attribute) -> Quality:
    # MORE STRICT IMPLEMENTATION ON SOME ATTRIBUTES
    if attribute in [Attribute.MOLD, Attribute.BENZENE, Attribute.FORMALDEHYDE, Attribute.CO2, Attribute.RADON]:
        if Quality.POOR in series.value_counts().index:
            return Quality.POOR
        elif Quality.FAIR in series.value_counts().index:
            return Quality.FAIR
        elif Quality.GOOD in series.value_counts().index:
            return Quality.GOOD
        elif Quality.EXCELLENT in series.value_counts().index:
            return Quality.EXCELLENT
        else:
            return np.nan
    else:
        freq = series.value_counts(normalize=True)
        return np.select(
                    [
                        (freq.get(Quality.EXCELLENT, 0) >= 0.94) & 
                        (freq.get(Quality.FAIR, 0) <= 0.01) &
                        (freq.get(Quality.POOR, 0) == 0),

                        (freq.get(Quality.GOOD, 0) > 0.05) &
                        (freq.get(Quality.FAIR, 0) <= 0.05) &
                        (freq.get(Quality.POOR, 0) <= 0.01),

                        (freq.get(Quality.FAIR, 0) > 0.05) &
                        (freq.get(Quality.POOR, 0) <= 0.05),

                    ],
                    [
                        Quality.EXCELLENT,
                        Quality.GOOD,
                        Quality.FAIR
                    ],
                    default=Quality.POOR
                ).item()

def categorise(building: Building, agg_level: str = 'room') -> Building:
    """
    Categorise the sensor data in the building object
    agg_level can be room or sensor depending on the level of aggregation
    """
    if agg_level not in ['room', 'sensor']:
        raise ValueError("agg_level must be either 'room' or 'sensor'")

    # TODO: IMPLEMENT BUILDING LEVEL CATEGORISATION
    
    # Iterate either over the rooms or sensors in the building
    if agg_level == 'room':
        iterable = building.rooms
    elif agg_level == 'sensor':
        iterable = [sensor for room in building.rooms for sensor in room.sensors]

    for itr in iterable:
        # Get the ventilation rates for the room
        if agg_level == 'room':
            rooms = [itr]
            ventilation_rates = [get_ventilation_rates(room=itr)]
        elif agg_level == 'sensor':
            # Check if rooms_id is empty
            if not itr.rooms_id:
                raise ValueError("Sensor must be assigned to a room")
            # Retrieve the room object from the building object using the room_id
            rooms = [room for room in building.rooms if room.id in itr.rooms_id]
            if not rooms:
                raise ValueError("Room not found in building")
            ventilation_rates = [get_ventilation_rates(room=room) for room in rooms]

        # Get the sensor into a dataframe
        measurements_df = itr.frame_data()

        # If itr.default_building_schedule is True or schedule is not None, then we can use the schedule to categorise the data
        schedule = building.schedule if itr.default_building_schedule else itr.schedule if itr.schedule else None

        if schedule is not None:
            # Convert the schedule to a pandas series
            schedule_series = convert_to_series(schedule)

            # Apply the closest schedule value to the measurements 
            schedule_series = schedule_series.reindex(measurements_df.index, method='nearest')
            # Add the schedule values to the measurements_df
            measurements_df = pd.concat([measurements_df, schedule_series], axis=1)
            measurements_df.columns = list(measurements_df.columns)[:-1] + ['Schedule']

            # Apply a boolean mask to the measurements_df to get the measurements for the row that have a schedule value > 0
            measurements_df = measurements_df[measurements_df['Schedule'] > 0]
            measurements_df.drop('Schedule', axis=1, inplace=True)

        measurements_df['Season'] = measurements_df.index.month.map(lambda x: 'Heating' if x in [10,11,12,1,2,3,4,5] else 'Non-heating')

        if Attribute.AIR_TEMPERATURE in measurements_df.columns:
            quality_series = measurements_df.apply(
                lambda row: (
                    Quality.EXCELLENT if 21 <= row[Attribute.AIR_TEMPERATURE] <= 23 else
                    Quality.GOOD if (20 <= row[Attribute.AIR_TEMPERATURE] < 21) or (23 < row[Attribute.AIR_TEMPERATURE] <= 24) else
                    Quality.FAIR if (19 <= row[Attribute.AIR_TEMPERATURE] < 20) or (24 < row[Attribute.AIR_TEMPERATURE] <= 25) else
                    Quality.POOR
                ) if row['Season'] == 'Heating' else (
                    Quality.EXCELLENT if 23.5 <= row[Attribute.AIR_TEMPERATURE] <= 25.5 else
                    Quality.GOOD if (23 <= row[Attribute.AIR_TEMPERATURE] < 23.5) or (25.5 < row[Attribute.AIR_TEMPERATURE] <= 26) else
                    Quality.FAIR if (22 <= row[Attribute.AIR_TEMPERATURE] < 23) or (26 < row[Attribute.AIR_TEMPERATURE] <= 27) else
                    Quality.POOR
                ),
                axis=1
            )

            itr.add_quality_for_attribute(attribute=Attribute.AIR_TEMPERATURE, quality=get_quality_for_attr(quality_series, Attribute.AIR_TEMPERATURE))

        if Attribute.RELATIVE_HUMIDITY in measurements_df.columns:
            if building.buildingType == BuildingType.OFFICE:
                quality_series = measurements_df.apply(
                    lambda row: (
                        Quality.EXCELLENT if 30 <= row[Attribute.RELATIVE_HUMIDITY] <= 50 else
                        Quality.GOOD if (25 <= row[Attribute.RELATIVE_HUMIDITY] < 30) or (50 < row[Attribute.RELATIVE_HUMIDITY] <= 60) else
                        Quality.FAIR if (20 <= row[Attribute.RELATIVE_HUMIDITY] < 25) or (60 < row[Attribute.RELATIVE_HUMIDITY] <= 70) else
                        Quality.POOR
                    ),
                    axis=1
                )

                itr.add_quality_for_attribute(attribute=Attribute.RELATIVE_HUMIDITY, quality=get_quality_for_attr(quality_series, Attribute.RELATIVE_HUMIDITY))

            elif building.buildingType == BuildingType.HOTEL:
                quality_series = measurements_df.apply(
                    lambda row: (
                        Quality.EXCELLENT if 30 <= row[Attribute.RELATIVE_HUMIDITY] <= 50 else
                        Quality.GOOD if (25 <= row[Attribute.RELATIVE_HUMIDITY] < 30) or (50 < row[Attribute.RELATIVE_HUMIDITY] <= 60) else
                        Quality.FAIR if 20 <= row[Attribute.RELATIVE_HUMIDITY] < 25 else
                        Quality.POOR
                    ),
                    axis=1
                )

                itr.add_quality_for_attribute(attribute=Attribute.RELATIVE_HUMIDITY, quality=get_quality_for_attr(quality_series, Attribute.RELATIVE_HUMIDITY))

            else:
                # Warn that the building type is not implemented
                print(f"Building type {building.buildingType} is not implemented for relative humidity categorisation")

        if Attribute.PM25 in measurements_df.columns:
            pm25_df = measurements_df[Attribute.PM25].copy()
            # If negative value, set to NaN
            pm25_df[pm25_df < 0] = np.nan
            # Set values below 1 to 0
            pm25_df[pm25_df < 1] = 0
            # Round to nearest whole number
            pm25_df = pm25_df.round()
            quality_series = pm25_df.map(
                lambda x: (
                    Quality.EXCELLENT if x < 10 else
                    Quality.FAIR if x > 25 else
                    Quality.POOR
                )
            )

            itr.add_quality_for_attribute(attribute=Attribute.PM25, quality=get_quality_for_attr(quality_series, Attribute.PM25))

        if Attribute.FORMALDEHYDE in measurements_df.columns:
            formaldehyde_df = measurements_df[Attribute.FORMALDEHYDE].copy()
            # If negative value, set to NaN
            formaldehyde_df[formaldehyde_df < 0] = np.nan
            # Set values below 1 to 0
            formaldehyde_df[formaldehyde_df < 1] = 0
            # Round to nearest whole number
            formaldehyde_df = formaldehyde_df.round()
            quality_series = formaldehyde_df.map(
                lambda x: (
                    Quality.EXCELLENT if x < 30 else
                    Quality.FAIR if x > 100 else
                    Quality.POOR
                ),
            )

            itr.add_quality_for_attribute(attribute=Attribute.FORMALDEHYDE, quality=get_quality_for_attr(quality_series, Attribute.FORMALDEHYDE))

        if Attribute.BENZENE in measurements_df.columns:
            benzene_df = measurements_df[Attribute.BENZENE].copy()
            # If negative value, set to NaN
            benzene_df[benzene_df < 0] = np.nan
            # Set values below 1 to 0
            benzene_df[benzene_df < 1] = 0
            # Round to nearest whole number
            benzene_df = benzene_df.round()
            quality_series = benzene_df.map(
                lambda x: (
                    Quality.EXCELLENT if x < 2 else
                    Quality.FAIR if x >= 5 else
                    Quality.POOR
                ),
            )

            itr.add_quality_for_attribute(attribute=Attribute.BENZENE, quality=get_quality_for_attr(quality_series, Attribute.BENZENE))

        if Attribute.RADON in measurements_df.columns:
            radon_df = measurements_df[Attribute.RADON].copy()
            # If negative value, set to NaN
            radon_df[radon_df < 0] = np.nan
            # Set values below 1 to 0
            radon_df[radon_df < 1] = 0
            # Round to nearest whole number
            radon_df = radon_df.round()
            quality_series = radon_df.map(
                lambda x: (
                    Quality.EXCELLENT if x < 100 else
                    Quality.FAIR if x > 300 else
                    Quality.POOR
                ),
            )

            itr.add_quality_for_attribute(attribute=Attribute.RADON, quality=get_quality_for_attr(quality_series, Attribute.RADON))

        if Attribute.VENTILATION_RATE in measurements_df.columns:
            ventilation_rates = ventilation_rates[0] # TODO: NEED TO IMPLEMENT FOR MULTIPLE ROOMS WITH SAME SENSOR
            quality_series = measurements_df.apply(
                lambda row: (
                    Quality.EXCELLENT if row[Attribute.VENTILATION_RATE] >= ventilation_rates[Quality.EXCELLENT] else
                    Quality.GOOD if (ventilation_rates[Quality.GOOD] <= row[Attribute.VENTILATION_RATE] < ventilation_rates[Quality.EXCELLENT]) else
                    Quality.FAIR if (ventilation_rates[Quality.FAIR] <= row[Attribute.VENTILATION_RATE] < ventilation_rates[Quality.GOOD]) else
                    Quality.POOR
                ),
                axis=1
            )

            itr.add_quality_for_attribute(attribute=Attribute.VENTILATION_RATE, quality=get_quality_for_attr(quality_series, Attribute.VENTILATION_RATE))

        if Attribute.MOLD in measurements_df.columns:
            quality_series = measurements_df.apply(
                lambda row: (
                    Quality.EXCELLENT if row[Attribute.MOLD] == Mold.NONE else
                    Quality.GOOD if row[Attribute.MOLD] == Mold.LOW else
                    Quality.FAIR if row[Attribute.MOLD] == Mold.MEDIUM else
                    Quality.POOR
                ),
                axis=1
            )

            itr.add_quality_for_attribute(attribute=Attribute.MOLD, quality=get_quality_for_attr(quality_series, Attribute.MOLD))

        if Attribute.DAYLIGHT_FACTOR in measurements_df.columns:
            quality_series = measurements_df.apply(
                lambda row: (
                    Quality.EXCELLENT if row[Attribute.DAYLIGHT_FACTOR] >= 5 else
                    Quality.GOOD if (3.3 <= row[Attribute.DAYLIGHT_FACTOR] < 5) else
                    Quality.FAIR if (2.0 <= row[Attribute.DAYLIGHT_FACTOR] < 3.3) else
                    Quality.POOR
                ),
                axis=1
            )

            itr.add_quality_for_attribute(attribute=Attribute.DAYLIGHT_FACTOR, quality=get_quality_for_attr(quality_series, Attribute.DAYLIGHT_FACTOR))

        # Get the 95th percentile of the CO2 values
        if Attribute.CO2 in measurements_df.columns:
            co2_df = measurements_df[Attribute.CO2].copy() # pd.Series
            co2_df = co2_df.dropna()
            # Replace values below const.OUTDOOR_CO2 with outdoor CO2 value
            co2_df[co2_df < const.OUTDOOR_CO2] = const.OUTDOOR_CO2
            # Get the 95th percentile of the CO2 values
            co2_95 = co2_df.quantile(0.95)
            
            # Categorise the CO2 values
            quality = np.select(
                [co2_95 < 800,
                (800 <= co2_95 <= 1000),
                (1000 < co2_95 <= 1200),
                co2_95 > 1200],
                [Quality.EXCELLENT, Quality.GOOD, Quality.FAIR, Quality.POOR],
                default=np.nan
            ).item()
            # Set the category to the itr
            itr.add_quality_for_attribute(attribute=Attribute.CO2, quality=quality)

        if Attribute.SOUND_PRESSURE_LEVEL in measurements_df.columns:
            spl_df = measurements_df[Attribute.SOUND_PRESSURE_LEVEL].copy()
            spl_df = spl_df.dropna()
            # Replace values below const.MIN_SPL with const.MIN_SPL
            spl_df[spl_df < const.MIN_SPL] = const.MIN_SPL
            # Get the 5th percentile of the SPL values
            spl_5 = spl_df.quantile(0.05)

            # Categorise the SPL values depending on the room type (SMALL_OFFICE, OPEN_PLAN, HOTEL_ROOM)
            IMPLEMENTED_ROOM_TYPES = const.NOISE_CATEGORIES.keys()
            for room in rooms:
                if room.roomType in IMPLEMENTED_ROOM_TYPES:
                    categories = const.NOISE_CATEGORIES[room.roomType]
                    quality = np.select(
                        [spl_5 < categories[Quality.EXCELLENT],
                        (categories[Quality.EXCELLENT] <= spl_5 < categories[Quality.GOOD]),
                        (categories[Quality.GOOD] <= spl_5 <= categories[Quality.FAIR]),
                        spl_5 > categories[Quality.FAIR]],
                        [Quality.EXCELLENT, Quality.GOOD, Quality.FAIR, Quality.POOR],
                        default=np.nan
                    ).item()
                    room.add_quality_for_attribute(attribute=Attribute.SOUND_PRESSURE_LEVEL, quality=quality)
                
                elif room.roomType not in IMPLEMENTED_ROOM_TYPES:
                    # Warn that the room type is not implemented
                    print(f"Room type {room.roomType} is not implemented for noise categorisation")

        if Attribute.ILLUMINANCE in measurements_df.columns:
            ill_df = measurements_df[Attribute.ILLUMINANCE].copy()
            ill_df = ill_df.dropna()
            # Replace values below const.MIN_ILLUMINANCE with const.MIN_ILLUMINANCE
            ill_df[ill_df < const.MIN_ILLUMINANCE] = const.MIN_ILLUMINANCE

            if building.buildingType == BuildingType.OFFICE:
                percentage = ill_df.between(300, 500).sum() / len(ill_df)
                quality = np.select(
                    [percentage > 0.6,
                        (0.4 <= percentage <= 0.6),
                        (0.1 <= percentage < 0.4),
                        percentage < 0.1],
                    [Quality.EXCELLENT, Quality.GOOD, Quality.FAIR, Quality.POOR],
                    default=np.nan
                ).item()
                itr.add_quality_for_attribute(attribute=Attribute.ILLUMINANCE, quality=quality)

            elif building.buildingType == BuildingType.HOTEL:
                percentage = (ill_df >= 100).sum() / len(ill_df)

                quality = np.select(
                    [percentage == 0,
                        (0 < percentage <= 0.5),
                        (0.5 < percentage <= 0.9),
                        percentage > 0.9],
                    [Quality.EXCELLENT, Quality.GOOD, Quality.FAIR, Quality.POOR],
                    default=np.nan
                ).item()
                itr.add_quality_for_attribute(attribute=Attribute.ILLUMINANCE, quality=quality)

            else:
                # Warn that the building type is not implemented
                print(f"Building type {building.buildingType} is not implemented for illuminance categorisation")

    if agg_level == 'sensor':
        # We will need to take the least performing quality for each sensor and attribute and set it to the room quality
        for room in building.rooms:
            quality_dict = {}
            for sensor in room.sensors:
                for attribute, quality in sensor.quality.items():
                    if attribute not in quality_dict:
                        quality_dict[attribute] = []
                    quality_dict[attribute].append(quality)
            
            for attribute, qualities in quality_dict.items():
                rat = min(qualities)
                room.add_quality_for_attribute(attribute=attribute, quality=rat)

    return building




import pandas as pd
from datetime import datetime

from typing import List

from .model import Building, Room, SensorData, DataPoint, Record, SensorAttribute, Attribute, Unit, RoomType, BuildingType, Mold

def parse_original_csv(file_path: str) -> List[Building]:
    df = pd.read_csv(file_path)

    building_names = df['building'].unique()

    buildings = []

    for building_name in building_names:
        df_building = df[df['building'] == building_name]
        # Occupancy is equal to the sum of the occupants in each room
        merge_df = df_building.groupby('room').agg({
            'occupancy': 'first',
            'floorArea': 'first',
            }).reset_index()
        occupancy = merge_df['occupancy'].sum().astype(float)
        floor_area = merge_df['floorArea'].sum().astype(float)

        rooms = df_building['room'].unique()

        if "buildingType" in df_building.columns:
            buildingType = df_building['buildingType'].unique()
            if len(buildingType) > 1:
                raise ValueError("Multiple building types found for the same building")
            else:
                buildingType = sanitize_name(buildingType[0])
                # Check is buildintType is valid
                if buildingType not in BuildingType.__members__:
                    raise ValueError(f"Invalid building type: {buildingType}")
                else:
                    buildingType = BuildingType[buildingType]
        else:
            buildingType = BuildingType.OFFICE
        
        building = Building(
            name=building_name,
            buildingType=buildingType,
            floorArea=floor_area,
            occupants=occupancy,
            rooms=[],
            coordinates="0,0"
        )

        for room in rooms:
            df_room = df_building[df_building['room'] == room]
            room_floor_area = df_room['floorArea'].unique()[0]
            room_occupants = df_room['occupancy'].unique()[0]
            room_type = df_room['roomType'].unique()
            if len(room_type) > 1:
                raise ValueError("Multiple room types found for the same room")
            else:
                room_type = sanitize_name(room_type[0])
                # Check if roomType is valid
                if room_type not in RoomType.__members__:
                    raise ValueError(f"Invalid room type: {room_type}")
                else:
                    room_type = RoomType[room_type]

            sensors = []
            # Create a dict of sensor name and sensor data
            dict_sensor = df_room.groupby('sensor').agg({
                'timestamp': list,
                'parameter': list,
                'value': list,
                'unit': list
            }).reset_index()

            dict_sensor = dict_sensor.to_dict(orient='records')

            for sensor in dict_sensor:
                # Create on sensor by sensor name
                sensor_name = sensor['sensor']
                # from the other columns create a df
                df_sensor = pd.DataFrame({
                    'timestamp': sensor['timestamp'],
                    'parameter': sensor['parameter'],
                    'value': sensor['value'],
                    'unit': sensor['unit']
                })
                # Group by timestamp
                df_sensor['timestamp'] = pd.to_datetime(df_sensor['timestamp'], dayfirst=True)
                df_sensor = df_sensor.groupby('timestamp').agg({
                    'parameter': list,
                    'value': list,
                    'unit': list
                }).reset_index()

                data_points = []
                attributes = set()
                for index, row in df_sensor.iterrows():
                    records = []
                    for i in range(len(row['parameter'])):
                        attributes.add(SensorAttribute(
                            attribute=Attribute[sanitize_name(row['parameter'][i])],
                            unit=Unit[sanitize_name(row['unit'][i])]
                        ))
                        records.append(Record(
                            sensorAttr=SensorAttribute(
                                attribute=Attribute[sanitize_name(row['parameter'][i])],
                                unit=Unit[sanitize_name(row['unit'][i])]
                            ),
                            value=row['value'][i] if Attribute[sanitize_name(row['parameter'][i])] != Attribute.MOLD else Mold[row['value'][i]]
                        ))
                    data_points.append(DataPoint(
                        timestamp=row['timestamp'],
                        records=records
                    ))

                sensors.append(SensorData(
                    sensorName=sensor_name,
                    attributes=list(attributes),
                    data=data_points
                ))

            room_obj = Room(
                name=room,
                floorArea=room_floor_area,
                occupants=room_occupants,
                roomType=room_type,
                sensors=sensors
            )

            building.add_room(room_obj)
        
        buildings.append(building)

    return buildings


def sanitize_name(name: str) -> str:
    return name.strip().replace(" ", "_").upper()


import pint

from .model import Attribute, Unit

# Initialize the pint UnitRegistry
ureg = pint.UnitRegistry()

# Add missing unit definitions
ureg.define('ug/m3 = microgram / meter ** 3')
ureg.define('Bq/m3 = becquerel / meter ** 3')
ureg.define('L/s = liter / second')

# Default units for each attribute
DEFAULT_UNITS = {
    Attribute.AIR_TEMPERATURE: Unit.CELSIUS,
    Attribute.RELATIVE_HUMIDITY: Unit.PERCENT,
    Attribute.ILLUMINANCE: Unit.LUX,
    Attribute.CO2: Unit.PPM,
    Attribute.PM25: Unit.MICROGRAM_PER_CUBIC_METER,
    Attribute.FORMALDEHYDE: Unit.MICROGRAM_PER_CUBIC_METER,
    Attribute.BENZENE: Unit.MICROGRAM_PER_CUBIC_METER,
    Attribute.RADON: Unit.BECQUEREL_PER_CUBIC_METER,
    Attribute.VENTILATION_RATE: Unit.LITERS_PER_SECOND,
    Attribute.SOUND_PRESSURE_LEVEL: Unit.DECIBEL,
    Attribute.DAYLIGHT_FACTOR: Unit.PERCENT,
    # Mold is not converted as it is categorical
}

def convert_unit(attribute: Attribute, value: float, from_unit: Unit, to_unit: Unit) -> float:
    # Use pint for unit conversion
    if from_unit == Unit.CELSIUS:
        # Convert Celsius to Kelvin for arithmetic operations
        value_kelvin = value + 273.15
        quantity = value_kelvin * ureg.kelvin
    elif from_unit == Unit.FAHRENHEIT:
        # Convert Fahrenheit to Kelvin for arithmetic operations
        value_kelvin = (value - 32) * 5/9 + 273.15
        quantity = value_kelvin * ureg.kelvin
    else:
        quantity = value * ureg(from_unit.value)

    if to_unit == Unit.CELSIUS:
        # Convert Kelvin back to Celsius
        converted_value = quantity.to(ureg.kelvin).magnitude - 273.15
    elif to_unit == Unit.FAHRENHEIT:
        # Convert Kelvin back to Fahrenheit
        converted_value = (quantity.to(ureg.kelvin).magnitude - 273.15) * 9/5 + 32
    else:
        converted_value = quantity.to(ureg(to_unit.value)).magnitude

    return converted_value

# Function to convert ureg units into the default unit for the passed ureg units only
def convert_to_default_units(value: float, from_unit: pint.Unit, attribute: Attribute) -> float:
    default_unit = DEFAULT_UNITS.get(attribute)
    if default_unit:
        if from_unit == ureg.celsius:
            # Convert Celsius to Kelvin for arithmetic operations
            value_kelvin = value + 273.15
            quantity = value_kelvin * ureg.kelvin
        elif from_unit == ureg.fahrenheit:
            # Convert Fahrenheit to Kelvin for arithmetic operations
            value_kelvin = (value - 32) * 5/9 + 273.15
            quantity = value_kelvin * ureg.kelvin
        else:
            quantity = value * from_unit

        if default_unit == Unit.CELSIUS:
            # Convert Kelvin back to Celsius
            converted_value = quantity.to(ureg.kelvin).magnitude - 273.15
        elif default_unit == Unit.FAHRENHEIT:
            # Convert Kelvin back to Fahrenheit
            converted_value = (quantity.to(ureg.kelvin).magnitude - 273.15) * 9/5 + 32
        else:
            converted_value = quantity.to(ureg(default_unit.value)).magnitude

        return converted_value
    return value


from enum import Enum
from datetime import datetime
from typing import Dict, Tuple
from pydantic import BaseModel
import pint

# Initialize the pint UnitRegistry
ureg = pint.UnitRegistry()

class Attribute(Enum):
    AIR_TEMPERATURE = "Air Temperature"
    RELATIVE_HUMIDITY = "Relative Humidity"
    ILLUMINANCE = "Illuminance"
    CO2 = "CO2"
    PM25 = "PM2.5"
    FORMALDEHYDE = "Formaldehyde"
    BENZENE = "Benzene"
    RADON = "Radon"
    VENTILATION_RATE = "Ventilation Rate"
    MOLD = "Mold"
    SOUND_PRESSURE_LEVEL = "Sound Pressure Level"
    DAYLIGHT_FACTOR = "Daylight Factor"

class Unit(Enum):
    CELSIUS = ureg.celsius
    FAHRENHEIT = ureg.fahrenheit
    PERCENT = ureg.percent
    LUX = ureg.lux
    PPM = ureg.ppm
    MICROGRAM_PER_CUBIC_METER = ureg.microgram / ureg.meter**3
    BECQUEREL_PER_CUBIC_METER = ureg.becquerel / ureg.meter**3
    LITERS_PER_SECOND = ureg.liter / ureg.second
    DECIBEL = ureg.decibel
    RATIO = ureg.dimensionless

# Default units for each attribute
DEFAULT_UNITS = {
    Attribute.AIR_TEMPERATURE: Unit.CELSIUS,
    Attribute.RELATIVE_HUMIDITY: Unit.PERCENT,
    Attribute.ILLUMINANCE: Unit.LUX,
    Attribute.CO2: Unit.PPM,
    Attribute.PM25: Unit.MICROGRAM_PER_CUBIC_METER,
    Attribute.FORMALDEHYDE: Unit.MICROGRAM_PER_CUBIC_METER,
    Attribute.BENZENE: Unit.MICROGRAM_PER_CUBIC_METER,
    Attribute.RADON: Unit.BECQUEREL_PER_CUBIC_METER,
    Attribute.VENTILATION_RATE: Unit.LITERS_PER_SECOND,
    Attribute.SOUND_PRESSURE_LEVEL: Unit.DECIBEL,
    Attribute.DAYLIGHT_FACTOR: Unit.PERCENT,
    # Mold is not converted as it is categorical
}

class SensorAttribute(BaseModel):
    attribute: Attribute
    unit: Unit

    def __hash__(self):
        return hash((self.attribute, self.unit))

    def __eq__(self, other):
        if isinstance(other, SensorAttribute):
            return self.attribute == other.attribute and self.unit == other.unit
        return False

class DataPoint(BaseModel):
    timestamp: datetime
    value: Dict[SensorAttribute, float]

    def convert_to_default_units(self) -> 'DataPoint':
        converted_values = {}
        for attr, value in self.value.items():
            default_unit = DEFAULT_UNITS.get(attr.attribute)
            if default_unit and default_unit != attr.unit:
                converted_value = self.convert_unit(attr.attribute, value, attr.unit, default_unit)
            else:
                converted_value = value
            converted_values[SensorAttribute(attribute=attr.attribute, unit=default_unit)] = converted_value
        return DataPoint(timestamp=self.timestamp, value=converted_values)

    @staticmethod
    def convert_unit(attribute: Attribute, value: float, from_unit: Unit, to_unit: Unit) -> float:
        # Use pint for unit conversion
        if from_unit == Unit.CELSIUS:
            # Convert Celsius to Kelvin for arithmetic operations
            value_kelvin = value + 273.15
            quantity = value_kelvin * ureg.kelvin
        elif from_unit == Unit.FAHRENHEIT:
            # Convert Fahrenheit to Kelvin for arithmetic operations
            value_kelvin = (value - 32) * 5/9 + 273.15
            quantity = value_kelvin * ureg.kelvin
        else:
            quantity = value * from_unit.value

        if to_unit == Unit.CELSIUS:
            # Convert Kelvin back to Celsius
            converted_value = quantity.to(ureg.kelvin).magnitude - 273.15
        elif to_unit == Unit.FAHRENHEIT:
            # Convert Kelvin back to Fahrenheit
            converted_value = (quantity.to(ureg.kelvin).magnitude - 273.15) * 9/5 + 32
        else:
            converted_value = quantity.to(to_unit.value).magnitude

        return converted_value

# Function to convert ureg units into the default unit for the passed ureg units only
def convert_to_default_units(value: float, from_unit: pint.Unit, attribute: Attribute) -> float:
    default_unit = DEFAULT_UNITS.get(attribute)
    if default_unit:
        if from_unit == ureg.celsius:
            # Convert Celsius to Kelvin for arithmetic operations
            value_kelvin = value + 273.15
            quantity = value_kelvin * ureg.kelvin
        elif from_unit == ureg.fahrenheit:
            # Convert Fahrenheit to Kelvin for arithmetic operations
            value_kelvin = (value - 32) * 5/9 + 273.15
            quantity = value_kelvin * ureg.kelvin
        else:
            quantity = value * from_unit

        if default_unit == Unit.CELSIUS:
            # Convert Kelvin back to Celsius
            converted_value = quantity.to(ureg.kelvin).magnitude - 273.15
        elif default_unit == Unit.FAHRENHEIT:
            # Convert Kelvin back to Fahrenheit
            converted_value = (quantity.to(ureg.kelvin).magnitude - 273.15) * 9/5 + 32
        else:
            converted_value = quantity.to(default_unit.value).magnitude

        return converted_value
    return value

# Example usage
# Create a DataPoint
data_point = DataPoint(
    timestamp=datetime.now(),
    value={
        SensorAttribute(attribute=Attribute.AIR_TEMPERATURE, unit=Unit.CELSIUS): 23,
        SensorAttribute(attribute=Attribute.AIR_TEMPERATURE, unit=Unit.FAHRENHEIT): 73.4,
        SensorAttribute(attribute=Attribute.RELATIVE_HUMIDITY, unit=Unit.PERCENT): 35,
        SensorAttribute(attribute=Attribute.PM25, unit=Unit.MICROGRAM_PER_CUBIC_METER): 6,
        SensorAttribute(attribute=Attribute.CO2, unit=Unit.PPM): 560,
        SensorAttribute(attribute=Attribute.FORMALDEHYDE, unit=Unit.MICROGRAM_PER_CUBIC_METER): 15,
        SensorAttribute(attribute=Attribute.BENZENE, unit=Unit.MICROGRAM_PER_CUBIC_METER): 2,
        SensorAttribute(attribute=Attribute.RADON, unit=Unit.BECQUEREL_PER_CUBIC_METER): 100,
        SensorAttribute(attribute=Attribute.VENTILATION_RATE, unit=Unit.LITERS_PER_SECOND): 10,
        SensorAttribute(attribute=Attribute.SOUND_PRESSURE_LEVEL, unit=Unit.DECIBEL): 45,
        SensorAttribute(attribute=Attribute.ILLUMINANCE, unit=Unit.LUX): 250,
        SensorAttribute(attribute=Attribute.DAYLIGHT_FACTOR, unit=Unit.PERCENT): 4.5,
        # Mold is not converted as it is categorical
    }
)

# Convert to default units
converted_data_point = data_point.convert_to_default_units()
print(converted_data_point.model_dump_json(indent=2))

# Example of converting a single value
value = 73.4
from_unit = ureg.fahrenheit
attribute = Attribute.AIR_TEMPERATURE
converted_value = convert_to_default_units(value, from_unit, attribute)
print(f"Converted value: {converted_value}")



from enum import Enum
from datetime import datetime
from typing import Dict, List, Optional, Union
from pydantic import BaseModel, Field, ConfigDict
from uuid import uuid4
import pandas as pd
import pickle
import numpy as np

from .schedule import YearSchedule

class Quality(Enum):
    EXCELLENT = 1
    GOOD = 2
    FAIR = 3
    POOR = 4

    @staticmethod
    def average(qualities: List['Quality']) -> 'Quality':
        return Quality(np.ceil(np.mean([q.value for q in qualities])).astype(int))
    
    # Add a method to compare the quality of two objects
    def __lt__(self, other: 'Quality') -> bool:
        return self.value < other.value
    
    def __le__(self, other: 'Quality') -> bool:
        return self.value <= other.value
    
    def __gt__(self, other: 'Quality') -> bool:
        return self.value > other.value
    
    def __ge__(self, other: 'Quality') -> bool:
        return self.value >= other.value
    
    def __eq__(self, other: 'Quality') -> bool:
        return self.value == other.value
    
    def __ne__(self, other: 'Quality') -> bool:
        return self.value != other.value
    
    def __hash__(self):
        return hash(self.value)
    
    def __str__(self):
        return f'{self.name}'

class BuildingType(Enum):
    RESIDENTIAL = "Residential"
    HOTEL = "Hotel"
    OFFICE = "Office"
    MIXED_USE = "Mixed Use"
    EDUCATIONAL = "Educational"

class Building(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, json_encoders={pd.DataFrame: lambda df: df.to_dict()})
    id: str = Field(default_factory=lambda: str(uuid4()))
    name: str
    floorArea: float
    occupants: int
    buildingType: BuildingType
    coordinates: str
    schedule: Optional['YearSchedule'] = None
    rooms: List['Room'] = []
    tail_rating: Optional['TAILRating'] = None

    def add_schedule(self, schedule: 'YearSchedule'):
        self.schedule = schedule
        for room in self.rooms:
            if room.schedule is None or room.default_building_schedule:
                room.assign_building_schedule()
            for sensor in room.sensors:
                if sensor.schedule is None or sensor.default_building_schedule:
                    sensor.assign_building_schedule()

    def add_room(self, room: 'Room'):
        if room not in self.rooms:
            self.rooms.append(room)
    
    def to_pickle(self, filename: str):
        with open(filename, 'wb') as f:
            pickle.dump(self, f)
            
    @staticmethod
    def from_pickle(filename: str) -> 'Building':
        with open(filename, 'rb') as f:
            return pickle.load(f)

class PollutionLevel(Enum):
    VERY_LOW_EMISSION = "Very Low Emission"
    LOW_EMISSION = "Low Emission"
    NON_LOW_EMISSION = "Non-Low Emission"

class TAIL(BaseModel):
    attributes: Dict['Attribute', Quality] = Field(default_factory=dict)
    indicators: Dict['Indicator', Quality] = Field(default_factory=dict)
    overall: Optional[Quality] = None

    def calculate_overall(self):
        self.overall = max((quality for quality in self.indicators.values() if quality is not None), default=None)

    def assign_indicator(self, indicator: 'Indicator', quality: Quality):
        self.indicators[indicator] = quality

    def assign_all_attributes(self, attributes: Dict['Attribute', Quality]):
        self.attributes = attributes

class Room(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, json_encoders={pd.DataFrame: lambda df: df.to_dict()})
    id: str = Field(default_factory=lambda: str(uuid4()))
    name: str
    floorArea: float
    occupants: int
    pollutionLevel: PollutionLevel = PollutionLevel.NON_LOW_EMISSION
    roomType: 'RoomType'
    schedule: Optional['YearSchedule'] = None
    default_building_schedule: bool = False
    sensors: Optional[List['SensorData']] = Field(default_factory=list)
    room_df: Optional[pd.DataFrame] = None
    tail: 'TAIL' = Field(default_factory=TAIL)

    def __init__(self, **data):
        super().__init__(**data)
        self.tail = TAIL()

    def assign_building_schedule(self):
        self.default_building_schedule = True

    def add_schedule(self, schedule: 'YearSchedule'):
        self.schedule = schedule

    def add_sensor(self, sensor: 'SensorData'):
        if sensor not in self.sensors:
            self.sensors.append(sensor)
            sensor.add_room_id(self)
    
    def frame_data(self):
        self.room_df = create_sensor_df(self.sensors)
        return self.room_df
    
    def add_quality_for_attribute(self, attribute: 'Attribute', quality: Quality):
        self.tail.attributes[attribute] = quality

class SensorData(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, json_encoders={pd.DataFrame: lambda df: df.to_dict()})
    id: str = Field(default_factory=lambda: str(uuid4()))
    sensorName: str
    attributes: List['SensorAttribute']
    data: List['DataPoint']
    schedule: Optional['YearSchedule'] = None
    default_building_schedule: bool = False
    sensor_df: Optional[pd.DataFrame] = None
    quality: Optional[Dict['Attribute', Quality]] = Field(default_factory=dict)
    rooms_id: Optional[List[str]] = Field(default_factory=list)

    def __init__(self, **data):
        super().__init__(**data)
        self.quality = {}

    def assign_building_schedule(self):
        self.default_building_schedule = True

    def add_schedule(self, schedule: 'YearSchedule'):
        self.schedule = schedule

    def add_room_id(self, room: Room):
        if room.id not in self.rooms_id:
            self.rooms_id.append(room.id)
        
    def frame_data(self):
        self.sensor_df = create_sensor_df([self])
        return self.sensor_df
    
    def add_quality_for_attribute(self, attribute: 'Attribute', quality: Quality, room: Room = None):
        self.quality[attribute] = quality
        if room is not None:
            room.add_quality_for_attribute(attribute, quality)

class RoomType(Enum):
    SMALL_OFFICE = "Single Office"
    OPEN_PLAN_OFFICE = "Open Plan Office"
    HOTEL_ROOM = "Hotel Room"
    CONFERENCE_ROOM = "Conference Room"
    CLASSROOM = "Classroom"
    LECTURE_HALL = "Lecture Hall"
    LIBRARY = "Library"
    AUDITORIUM = "Auditorium"
    GYM = "Gym"
    RESTAURANT = "Restaurant"
    KITCHEN = "Kitchen"

class Attribute(Enum):
    AIR_TEMPERATURE = "Air Temperature"
    RELATIVE_HUMIDITY = "Relative Humidity"
    ILLUMINANCE = "Illuminance"
    CO2 = "CO2"
    PM25 = "PM2.5"
    FORMALDEHYDE = "Formaldehyde"
    BENZENE = "Benzene"
    RADON = "Radon"
    VENTILATION_RATE = "Ventilation Rate"
    MOLD = "Mold"
    SOUND_PRESSURE_LEVEL = "Sound Pressure Level"
    DAYLIGHT_FACTOR = "Daylight Factor"

    @classmethod
    def list(cls):
        return list(cls)

class Mold(Enum):
    NONE = "None"
    LOW = "Low"
    MEDIUM = "Medium"
    HIGH = "High"

class Unit(Enum):
    CELSIUS = "celsius"
    FAHRENHEIT = "fahrenheit"
    PERCENT = "percent"
    LUX = "lux"
    PPM = "ppm"
    MICROGRAM_PER_CUBIC_METER = "ug/m3"
    BECQUEREL_PER_CUBIC_METER = "Bq/m3"
    LITERS_PER_SECOND = "L/s"
    DECIBEL = "dB(A)"
    RATIO = "ratio"
    CAT = "cat"

class SensorAttribute(BaseModel):
    attribute: Attribute
    unit: Unit
    
    def __hash__(self):
        return hash((self.attribute, self.unit))

    def __eq__(self, other):
        if isinstance(other, SensorAttribute):
            return self.attribute == other.attribute and self.unit == other.unit
        return False

class Record(BaseModel):
    sensorAttr: SensorAttribute
    value: Union[float, Mold, int]

    def is_valid(self) -> bool:
        if self.sensorAttr.attribute == Attribute.MOLD:
            return isinstance(self.value, Mold)
        return isinstance(self.value, (float, int))

class DataPoint(BaseModel):
    timestamp: datetime
    records: List[Record]

    def convert_to_default_units(self) -> 'DataPoint':
        from .units import convert_unit
        converted_values = {}
        for record in self.records:
            if record.sensorAttr.attribute in DEFAULT_UNITS:
                converted_values[record.sensorAttr] = convert_unit(record.value, record.sensorAttr.unit, DEFAULT_UNITS[record.sensorAttr.attribute])
            else:
                converted_values[record.sensorAttr] = record.value
        return DataPoint(timestamp=self.timestamp, records=[Record(sensorAttr=attr, value=val) for attr, val in converted_values.items()])

DEFAULT_UNITS = {
    Attribute.AIR_TEMPERATURE: Unit.CELSIUS,
    Attribute.RELATIVE_HUMIDITY: Unit.PERCENT,
    Attribute.ILLUMINANCE: Unit.LUX,
    Attribute.CO2: Unit.PPM,
    Attribute.PM25: Unit.MICROGRAM_PER_CUBIC_METER,
    Attribute.FORMALDEHYDE: Unit.MICROGRAM_PER_CUBIC_METER,
    Attribute.BENZENE: Unit.MICROGRAM_PER_CUBIC_METER,
    Attribute.RADON: Unit.BECQUEREL_PER_CUBIC_METER,
    Attribute.VENTILATION_RATE: Unit.LITERS_PER_SECOND,
    Attribute.SOUND_PRESSURE_LEVEL: Unit.DECIBEL,
    Attribute.DAYLIGHT_FACTOR: Unit.PERCENT,
}

class Indicator(Enum):
    THERMAL = "Thermal"
    ACOUSTIC = "Acoustic"
    INDOOR_AIR_QUALITY = "Indoor Air Quality"
    LIGHTING = "Lighting"

    @classmethod
    def list(cls):
        return list(cls)

class TAIL(BaseModel):
    attributes: Dict[Attribute, Quality] = Field(default_factory=dict)
    indicators: Dict[Indicator, Quality] = Field(default_factory=dict)
    overall: Optional[Quality] = None

    def calculate_overall(self):
        self.overall = max((quality for quality in self.indicators.values() if quality is not None), default=None)

    def assign_indicator(self, indicator: Indicator, quality: Quality):
        self.indicators[indicator] = quality

    def assign_all_attributes(self, attributes: Dict[Attribute, Quality]):
        self.attributes = attributes

class TAILRating(BaseModel):
    id: str = Field(default_factory=lambda: str(uuid4()))
    from_date: datetime
    to_date: datetime
    tail: Optional[TAIL]
    name: str
    building_id: str
    rooms_id: Optional[List[str]] = Field(default_factory=list)
    sensors_id: Optional[List[str]] = Field(default_factory=list)
    valid: bool = True
    valid_until: Optional[datetime] = None

    def __init__(self, **data):
        super().__init__(**data)
        self.valid_until = self.valid_until or self.from_date.replace(year=self.from_date.year + 1)
        
    def add_room_id(self, room: Room):
        if room.id not in self.rooms_id:
            self.rooms_id.append(room.id)

    def add_sensor_id(self, sensor: SensorData):
        if sensor.id not in self.sensors_id:
            self.sensors_id.append(sensor.id)

    def invalidate(self):
        self.valid = False
        self.valid_until = datetime.now()

def create_sensor_df(sensor_data: List[SensorData]) -> pd.DataFrame:
    data = []
    for sensor in sensor_data:
        for dataPoint in sensor.data:
            row = {'timestamp': dataPoint.timestamp}
            for record in dataPoint.records:
                row[record.sensorAttr.attribute] = record.value
            data.append(row)

    df = pd.DataFrame(data)
    df.set_index('timestamp', inplace=True)
    df.index = df.index.round('s')
    df = df.groupby(df.index).agg(lambda x: x.mean() if x.dtype.kind in 'biufc' else (x.mode().iat[0] if not x.mode().empty else np.nan))
    df.sort_index(inplace=True)

    return df



import unittest
from TAIL.model import Quality

class TestQualityEnum(unittest.TestCase):

    def test_equality(self):
        self.assertTrue(Quality.EXCELLENT == Quality.EXCELLENT)
        self.assertFalse(Quality.EXCELLENT == Quality.GOOD)

    def test_inequality(self):
        self.assertTrue(Quality.EXCELLENT != Quality.GOOD)
        self.assertFalse(Quality.EXCELLENT != Quality.EXCELLENT)

    def test_less_than(self):
        self.assertTrue(Quality.GOOD < Quality.EXCELLENT)
        self.assertFalse(Quality.EXCELLENT < Quality.GOOD)

    def test_less_than_or_equal(self):
        self.assertTrue(Quality.GOOD <= Quality.EXCELLENT)
        self.assertTrue(Quality.EXCELLENT <= Quality.EXCELLENT)
        self.assertFalse(Quality.EXCELLENT <= Quality.GOOD)

    def test_greater_than(self):
        self.assertTrue(Quality.EXCELLENT > Quality.GOOD)
        self.assertFalse(Quality.GOOD > Quality.EXCELLENT)

    def test_greater_than_or_equal(self):
        self.assertTrue(Quality.EXCELLENT >= Quality.GOOD)
        self.assertTrue(Quality.EXCELLENT >= Quality.EXCELLENT)
        self.assertFalse(Quality.GOOD >= Quality.EXCELLENT)

if __name__ == '__main__':
    unittest.main()


from setuptools import setup, find_packages

with open('requirements.txt') as f:
    required = f.read().splitlines()

setup(
    name='tail-rating-py',
    version='0.1.2',
    packages=find_packages(),
    install_requires=required,
    author='Bruno Adam',
    author_email='bruno.adam@pm.me',
    description='A Python package designed to assess TAIL (Thermal, Acoustic, Indoor Air Quality, and Lighting) for buildings.',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url='',
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.7',
    license='GNU General Public License v3 (GPLv3)',
    keywords='tail thermal acoustic indoor air quality lighting',
    # The folder is called TAIL, but the package is called TAILpy
    package_dir={'TAIL': 'TAIL'},
)


from .model import RoomType, Quality, BuildingType 

OUTDOOR_CO2 = 420 # ppm
MIN_SPL = 20 # dB(A) - Minimum sound pressure level based on the measurement device

NOISE_CATEGORIES = {
    RoomType.SMALL_OFFICE: {
        Quality.EXCELLENT: 30,
        Quality.GOOD: 35,
        Quality.FAIR: 40
    },
    RoomType.OPEN_PLAN_OFFICE: {
        Quality.EXCELLENT: 35,
        Quality.GOOD: 40,
        Quality.FAIR: 45
    },
    RoomType.HOTEL_ROOM: {
        Quality.EXCELLENT: 25,
        Quality.GOOD: 30,
        Quality.FAIR: 35
    }
}

MIN_ILLUMINANCE = 30 # lux



from datetime import datetime, timedelta
from .model import Building, Room, SensorData, DataPoint, SensorAttribute, Attribute, Unit, BuildingType, Record, RoomType, Mold

import random

from .schedule import create_basic_schedule

# Create some fake data points
def create_fake_data_points(attributes):
    data_points = []
    from_date = datetime(2025, 1, 1, 0, 0)
    to_date = datetime(2025, 1, 31, 23, 30)
    timestep = timedelta(minutes=30)
    
    # Create a list of timestamps
    timestamps = [from_date + i * timestep for i in range(int((to_date - from_date) / timestep))]   

    # Create a list of records for each timestamp, create a random value for each sensor attribute
    for timestamp in timestamps:
        records = [Record(sensorAttr=attribute, value=random.uniform(0, 100)) for attribute in attributes if attribute.attribute != Attribute.MOLD]
        
        # For SensorAttribute that contains attribute Mold, we need to assign a random value from the Mold enum:
        if any(attribute.attribute == Attribute.MOLD for attribute in attributes):
            records.append(Record(sensorAttr=SensorAttribute(attribute=Attribute.MOLD, unit=Unit.CAT), value=Mold(random.choice([Mold.NONE, Mold.LOW, Mold.MEDIUM, Mold.HIGH]))))

        
        data_points.append(DataPoint(timestamp=timestamp, records=records))

    return data_points

def create_building_model() -> Building:
    # Create sensor attributes
    sensor_attributes_1 = [
        SensorAttribute(attribute=Attribute.AIR_TEMPERATURE, unit=Unit.CELSIUS),
        SensorAttribute(attribute=Attribute.RELATIVE_HUMIDITY, unit=Unit.PERCENT),
    ]

    sensor_one_time_measurements = [
        SensorAttribute(attribute=Attribute.ILLUMINANCE, unit=Unit.LUX),
        SensorAttribute(attribute=Attribute.SOUND_PRESSURE_LEVEL, unit=Unit.DECIBEL),
        SensorAttribute(attribute=Attribute.DAYLIGHT_FACTOR, unit=Unit.PERCENT),
        SensorAttribute(attribute=Attribute.MOLD, unit=Unit.CAT)
    ]

    sensor_attributes_2 = [
        SensorAttribute(attribute=Attribute.CO2, unit=Unit.PPM),
        SensorAttribute(attribute=Attribute.PM25, unit=Unit.MICROGRAM_PER_CUBIC_METER)
    ]

    # Create sensors
    sensor_1 = SensorData(sensorName="Sensor 1", attributes=sensor_attributes_1, data=create_fake_data_points(sensor_attributes_1))
    sensor_2 = SensorData(sensorName="Sensor 2", attributes=sensor_attributes_2, data=create_fake_data_points(sensor_attributes_2))

    sensor_3 = SensorData(sensorName="Sensor 3", attributes=sensor_attributes_1, data=create_fake_data_points(sensor_attributes_1))
    sensor_4 = SensorData(sensorName="Sensor 4", attributes=sensor_attributes_2, data=create_fake_data_points(sensor_attributes_2))

    sensor_5 = SensorData(sensorName="Sensor 5", attributes=sensor_one_time_measurements, data=create_fake_data_points(sensor_one_time_measurements))

    # Create rooms
    room_1 = Room(name="Room 1", floorArea=50.0, occupants=5, roomType=RoomType.SMALL_OFFICE)
    room_2 = Room(name="Room 2", floorArea=75.0, occupants=10, roomType=RoomType.HOTEL_ROOM, sensors=[sensor_3, sensor_4, sensor_5])

    room_1.add_sensor(sensor_1)
    room_1.add_sensor(sensor_2)
    room_1.add_sensor(sensor_5)

    # Save the data using pickle
    building = Building(name="Building 0", floorArea=500.0, occupants=50, buildingType=BuildingType.OFFICE, coordinates="0,0", rooms=[room_1, room_2])

    # Create a schedule for the building
    schedule = create_basic_schedule(2025, "DK")
    building.add_schedule(schedule)

    return building



from .model import BuildingType, Quality, Room, PollutionLevel, RoomType

Q_p = {
    Quality.EXCELLENT: 10,
    Quality.GOOD: 7,
    Quality.FAIR: 4,
    Quality.POOR: 2.5
}

Q_B = {
    PollutionLevel.VERY_LOW_EMISSION: {
        Quality.EXCELLENT: 0.5,
        Quality.GOOD: 0.35,
        Quality.FAIR: 0.2,
        Quality.POOR: 0.15
    },
    PollutionLevel.LOW_EMISSION: {
        Quality.EXCELLENT: 1.0,
        Quality.GOOD: 0.7,
        Quality.FAIR: 0.4,
        Quality.POOR: 0.3
    },
    PollutionLevel.NON_LOW_EMISSION: {
        Quality.EXCELLENT: 2.0,
        Quality.GOOD: 1.4,
        Quality.FAIR: 0.8,
        Quality.POOR: 0.6
    },
}

def get_ventilation_rates(room: Room):
    pollution_level = room.pollutionLevel
    occupants = room.occupants
    floorArea = room.floorArea

    # Return a dictionary with the ventilation rates for each quality level calculated using the formula Q = Q_p * occupants + Q_B * floorArea

    return {
        quality: Q_p[quality] * occupants + Q_B[pollution_level][quality] * floorArea
        for quality in Quality
    }